# Copyright (c) Acconeer AB, 2022
# All rights reserved

ACCONEER_VID = 0x0483
ACCONEER_XC120_BOARD_PROTOCOL_PID = 0xA42C
ACCONEER_XC120_EXPLORATION_SERVER_PID = 0xA42D
ACCONEER_XC120_BOOTLOADER_PID = 0xA41D
